﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits.BanditPolicies {
  // Powell, Approximate Dynamic Programming, section 12.3.5, page 466, 
  public class IntervalEstimationPolicy : IBanditPolicy {
    public int SelectAction(Random random, IEnumerable<IBanditPolicyActionInfo> actionInfos) {
      var myActionInfos = actionInfos.OfType<MeanAndVariancePolicyActionInfo>();
      double bestQ = double.NegativeInfinity;
      int totalTries = myActionInfos.Sum(a => a.Tries);

      var bestActions = new List<int>();
      int aIdx = -1;
      foreach (var aInfo in myActionInfos) {
        aIdx++;
        double q;
        if (aInfo.Tries == 0) {
          q = double.PositiveInfinity;
        } else {
          double sigmaSqr = aInfo.RewardVariance * 1.0 / aInfo.Tries;
          q = aInfo.SumReward / aInfo.Tries + 3.0 * Math.Sqrt(sigmaSqr); // 3.0 for around 95th percentile
        }
        if (q > bestQ) {
          bestQ = q;
          bestActions.Clear();
          bestActions.Add(aIdx);
        } else if (q.IsAlmost(bestQ)) {
          bestActions.Add(aIdx);
        }
      }
      Debug.Assert(bestActions.Any());
      return bestActions.SelectRandom(random);
    }

    public IBanditPolicyActionInfo CreateActionInfo() {
      return new MeanAndVariancePolicyActionInfo();
    }
    public override string ToString() {
      return "IntervalEstimationPolicy(Powell)";
    }
  }
}
